@@ -3,7 +3,7 @@
 
  * Copyright (C) 2004  Andrew Reading
  * Copyright (C) 2005, 2006  Dan Walma
- * Copyright (C) 2008  Sylvain Beucler
+ * Copyright (C) 2008, 2014, 2018  Sylvain Beucler
 
  * This file is part of GNU FreeDink
 
@@ -31,6 +31,7 @@
 #include <wx/intl.h>
 #include <wx/log.h>
 #include <wx/filename.h>
+#include <wx/tokenzr.h>
 
 #include <math.h>
 #include <fstream>
@@ -41,6 +42,47 @@
 
 #include "IOUtils.hpp"
 
+#if defined _WIN32 || defined __WIN32__ || defined __CYGWIN__
+  #define fopen(file, mode) _wfopen(file, _T(mode))
+  // can't do the same with 'stat' as it's already a macro in mingw32
+
+  // See doc/unicode.txt
+  #include <ext/stdio_filebuf.h>
+  std::istream* istream_open_unicode(const wchar_t* filename) {
+    int fd = _wopen(filename, _O_RDONLY|O_BINARY);
+    __gnu_cxx::stdio_filebuf<char>* filebuf =
+      new __gnu_cxx::stdio_filebuf<char>(fd, std::ios::in|std::ios::binary);
+    return new std::istream(filebuf);
+  }
+  std::ostream* ostream_open_unicode(const wchar_t* filename) {
+    int fd = _wopen(filename, _O_CREAT|_O_WRONLY|_O_BINARY);
+    __gnu_cxx::stdio_filebuf<char>* filebuf =
+      new __gnu_cxx::stdio_filebuf<char>(fd, std::ios::out|std::ios::binary);
+    return new std::ostream(filebuf);
+  }
+  void istream_close_unicode(std::istream* s) {
+    delete s->rdbuf();
+    delete s;
+  }
+  void ostream_close_unicode(std::ostream* s) {
+    delete s->rdbuf();
+    delete s;
+  }
+#else
+  std::istream* istream_open_unicode(const char* filename) {
+    return new std::ifstream(filename, std::ios::in|std::ios::binary);
+  }
+  std::ostream* ostream_open_unicode(const char* filename) {
+    return new std::ofstream(filename, std::ios::out|std::ios::binary);
+  }
+  void istream_close_unicode(std::istream* s) {
+    delete s;
+  }
+  void ostream_close_unicode(std::ostream* s) {
+    delete s;
+  }
+#endif
+
 struct fileinfo
 {
   wxString fullpath;
@@ -74,11 +116,11 @@ bool Tar::Create(const wxString& dmod_folder, double *compression_ratio, wxProgr
   wxString strCwd = ::wxGetCwd();
 
   // Actual archiving
-  chdir(strCompressDir.fn_str());
+  ::wxSetWorkingDirectory(strCompressDir);
   bool result = CreateReal(dmod_folder, compression_ratio, aProgressDialog);
 
   // Back to where we were
-  chdir(strCwd.fn_str());
+  ::wxSetWorkingDirectory(strCwd);
 
   return result;
 }
@@ -100,8 +142,13 @@ bool Tar::CreateReal(const wxString& dmod_folder, double *compression_ratio, wxP
   for (unsigned int i = 0; i < wxasFileList.GetCount(); ++i)
     {
       fileinfos[i].fullpath = strCompressDir + _T("/") + wxasFileList.Item(i);
+#if defined _WIN32 || defined __WIN32__ || defined __CYGWIN__
+      struct _stat sb;
+      if (_wstat(fileinfos[i].fullpath.fn_str(), &sb) < 0)
+#else
       struct stat sb;
       if (stat(fileinfos[i].fullpath.fn_str(), &sb) < 0)
+#endif
 	{
 	  perror("stat");
 	  fileinfos[i].mode = 0; // invalid, to skip
@@ -115,7 +162,8 @@ bool Tar::CreateReal(const wxString& dmod_folder, double *compression_ratio, wxP
 #ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
       fileinfos[i].blksize = sb.st_blksize;
 #else
-      fileinfos[i].blksize = BUFSIZ;
+      // fileinfos[i].blksize = BUFSIZ;  // BUFSIZ=512 under woe, performance issues reported
+      fileinfos[i].blksize = 8192;
 #endif
 
       total_file_data += sb.st_size;
@@ -126,8 +174,8 @@ bool Tar::CreateReal(const wxString& dmod_folder, double *compression_ratio, wxP
   FILE* out = fopen(mFilePath.fn_str(), "wb");
   if (!out)
     {
-      ::wxLogError(_("Error: Could not open tar file '%s' for bzip compression."), mFilePath.c_str());
-      return _("ERROR");
+      wxLogError(_("Error: Could not open tar file '%s' for bzip compression."), mFilePath.c_str());
+      return false;
     }
   
   /* libbz2 init */
@@ -148,7 +196,7 @@ bool Tar::CreateReal(const wxString& dmod_folder, double *compression_ratio, wxP
   
   if (iErr != BZ_OK)
     {
-      ::wxLogError(_("Error: Could not initialize compression method!"
+      wxLogError(_("Error: Could not initialize compression method!"
 		     "  "
 		     "Will not generate a correct .dmod file."
 		     "  "
@@ -181,7 +229,7 @@ bool Tar::CreateReal(const wxString& dmod_folder, double *compression_ratio, wxP
       FILE *in = fopen(fileinfos[i].fullpath.fn_str(), "rb");
       if (in == NULL)
         {
-	  ::wxLogFatalError(_("Error: File '%s' not found!  Cannot archive file."),
+	  wxLogFatalError(_("Error: File '%s' not found!  Cannot archive file."),
 			    fileinfos[i].fullpath.c_str());
 	  throw;
         }
@@ -365,7 +413,7 @@ int Tar::ReadHeaders( void )
       memset(&lHeader, 0, sizeof(TarHeader));
       memset(&lRecord, 0, sizeof(TarRecord));
       
-      // Read the data. Dont load the whole header to a struct, you
+      // Read the data. Don't load the whole header to a struct, you
       // never know what the memory alignment will be, especially in
       // these 32->64bits days
       fread((char*)&lHeader.Name, 100, 1, in);
@@ -417,7 +465,15 @@ int Tar::ReadHeaders( void )
       wxString lPath(lRecord.Name, wxConvUTF8);
       if (mInstalledDmodDirectory.Length() == 0)
         {
-	  mInstalledDmodDirectory = lPath.SubString( 0, lPath.Find( '/' ) ).c_str();
+	  // Security: ensure the D-Mod directory is non-empty
+	  wxString firstDir = GetFirstDir(lPath);
+	  if (firstDir.IsSameAs("", true) || firstDir.IsSameAs("..", true) || firstDir.IsSameAs("dink", true))
+            {
+	      wxLogError(_("Error: invalid D-Mod directory.  Stopping."));
+              return 1;
+            }
+          mInstalledDmodDirectory = firstDir;
+
 	  lDmodDizPath = mInstalledDmodDirectory + _T("dmod.diz");
 	  lDmodDizPath.LowerCase();
         }
@@ -460,20 +516,15 @@ int Tar::ReadHeaders( void )
 int Tar::Extract(wxString destdir, wxProgressDialog* aProgressDialog)
 {
     wxString strBuf;
-    std::fstream f_In;
     int lError = 0;
 
-    // Remember current directory
-    wxString strCwd = ::wxGetCwd();
-
-
     // Open the file here so it doesn't error after changing.
-    f_In.open(mFilePath.fn_str(), std::ios::in | std::ios::binary);
-    
-    if(!f_In.good())
+    std::istream* f_In = istream_open_unicode(mFilePath.fn_str());
+
+    if(!f_In->good())
     {
       lError = 1;
-      ::wxLogFatalError(_("Error: File '%s' not found!  Cannot extract data."), mFilePath.c_str());
+      wxLogFatalError(_("Error: File '%s' not found!  Cannot extract data."), mFilePath.c_str());
       throw;
     }
     
@@ -483,46 +534,36 @@ int Tar::Extract(wxString destdir, wxProgressDialog* aProgressDialog)
       wxFileName::Mkdir(destdir, 0777, wxPATH_MKDIR_FULL); // 0777 minus umask
     if (!::wxDirExists(destdir))
       {
-	::wxLogFatalError(_("Error: Cannot create directory '%s'.  Cannot extract data."), destdir.c_str());
+	wxLogFatalError(_("Error: Cannot create directory '%s'.  Cannot extract data."), destdir.c_str());
 	throw;
       }
-    // Move to the directory.
-    chdir(destdir.fn_str());
 
     // Put the data in the directories.
     if (ExtractData(f_In, destdir, aProgressDialog) != 0)
     {
         lError = 1;
     }
-    f_In.close();
-    f_In.clear();
+    istream_close_unicode(f_In);
 
-
-    // We're done.  Move back.
-    chdir(strCwd.fn_str());
-    
     return lError;
 }
 
-int Tar::ExtractData(std::fstream& aTarStreamIn, wxString destdir, wxProgressDialog* aProgressDialog)
+int Tar::ExtractData(std::istream* aTarStreamIn, wxString destdir, wxProgressDialog* aProgressDialog)
 {
-    std::fstream lStreamOut;
     int lError = 0;
 
     unsigned long lTotalBytes = 0;
     unsigned long lTotalBytesRead = 0;
 
     // Get the file size
-    aTarStreamIn.seekg(0, std::ios::end);
-    unsigned long lEnd = static_cast<unsigned long>(aTarStreamIn.tellg());
-    aTarStreamIn.seekg(0, std::ios::beg);
-    lTotalBytes = lEnd - static_cast<unsigned long>(aTarStreamIn.tellg());
-
-    // Move into the extract dir.
-    wxString lPreviousWorkingDirectory(::wxGetCwd());
-    chdir(destdir.fn_str());
+    aTarStreamIn->seekg(0, std::ios::end);
+    unsigned long lEnd = static_cast<unsigned long>(aTarStreamIn->tellg());
+    aTarStreamIn->seekg(0, std::ios::beg);
+    lTotalBytes = lEnd - static_cast<unsigned long>(aTarStreamIn->tellg());
 
     // Extract the files.
+    int ebufsiz = 8192;
+    char buffer[ebufsiz];
     for (unsigned int lTarRecordIndex = 0; lTarRecordIndex < TarRecords.size(); ++lTarRecordIndex)
     {
 		TarRecord lCurrentTarRecord = TarRecords.at(lTarRecordIndex);
@@ -532,7 +573,16 @@ int Tar::ExtractData(std::fstream& aTarStreamIn, wxString destdir, wxProgressDia
 		    /* Attempt convertion from latin-1 if not valid UTF-8 */
 		    lCurrentFilePath = wxString(lCurrentTarRecord.Name, wxConvISO8859_1);
 		  }
-		wxString lCurrentDirectory(lCurrentFilePath.substr(0, lCurrentFilePath.find_last_of('/')));
+        // Security: check if archive tries to jump out of destination directory
+        if (IsPathInsecure(lCurrentFilePath))
+        {
+            wxLogError(_("Error: Insecure filename: '%s'.  Stopping."), lCurrentFilePath);
+            lError = 1;
+            break;
+        }
+        // Security: ensure full, non-relative path, under destdir/
+        lCurrentFilePath = destdir + wxFileName::GetPathSeparator() + lCurrentFilePath;
+        wxString lCurrentDirectory = lCurrentFilePath.substr(0, lCurrentFilePath.find_last_of("/\\"));
 
         // Odd bad file problem...
 	if (lCurrentFilePath.compare(_T("\xFF")) == 0) // ""
@@ -545,42 +595,44 @@ int Tar::ExtractData(std::fstream& aTarStreamIn, wxString destdir, wxProgressDia
             }
         }
         else
-        {
-		    // Create the directory if it doesn't exist
-		    if (::wxDirExists(lCurrentDirectory) == false)
-		    {
-		      // Use 0777 - the umask will remove group and/or
-		      // other write access as necessary. Don't forget
-		      // to prefix with a '0' for octal mode.
-		      wxFileName::Mkdir(lCurrentDirectory, 0777, wxPATH_MKDIR_FULL);
-		    }
-
+	{
+	    // Create the directory if it doesn't exist
+	    if (::wxDirExists(lCurrentDirectory) == false)
+	    {
+		// Use 0777 - the umask will remove group and/or
+		// other write access as necessary. Don't forget
+		// to prefix with a '0' for octal mode.
+		wxFileName::Mkdir(lCurrentDirectory, 0777, wxPATH_MKDIR_FULL);
+	    }
+	    
             // Only write the file if it is a file... some files are stored as directories
             char lLastCharacter = lCurrentFilePath[lCurrentFilePath.length() - 1];
             if (lLastCharacter != '\\' && lLastCharacter != '/')
             {
                 // Open the file.
-                lStreamOut.open(lCurrentFilePath.fn_str(), std::ios::out | std::ios::binary);
-                if (lStreamOut.good() == false)
+                std::ostream* lStreamOut = ostream_open_unicode(lCurrentFilePath);
+		if (!lStreamOut->good())
                 {
-                    lStreamOut.close();
-                    lStreamOut.clear();
-		    ::wxLogError(_("Error: Improperly archived file '%s'.  Skipping."), lCurrentFilePath.c_str());
+		    wxLogError(_("Error: Improperly archived file '%s'.  Skipping."), lCurrentFilePath);
                     lError = 1;
                     continue;
                 }
                 
-                // Write it (one byte at a time?  Seems slow...)
-		        aTarStreamIn.seekg( lCurrentTarRecord.iFilePosBegin, std::ios::beg );
-                for ( int j = 0; j < lCurrentTarRecord.iFileSize; ++j )
-                {
-                    lStreamOut.put( aTarStreamIn.get() );
-                }
+		aTarStreamIn->seekg(lCurrentTarRecord.iFilePosBegin, std::ios::beg);
+		long remaining = lCurrentTarRecord.iFileSize;
+		while (remaining > 0)
+		  {
+		    int bufsiz = remaining;
+		    if (bufsiz > ebufsiz) bufsiz = ebufsiz;
+		    aTarStreamIn->read(buffer, bufsiz);
+		    lStreamOut->write(buffer, bufsiz);
+		    remaining -= bufsiz;
+		  }
                 lTotalBytesRead += lCurrentTarRecord.iFileSize;
                 
                 // Close up.
-                lStreamOut.close();
-                lStreamOut.clear();
+                lStreamOut->flush();
+		ostream_close_unicode(lStreamOut);
             }
             
             // Update progress
@@ -597,7 +649,6 @@ int Tar::ExtractData(std::fstream& aTarStreamIn, wxString destdir, wxProgressDia
         }
     }
     aProgressDialog->Update(100, _("Done."));
-    chdir(lPreviousWorkingDirectory.fn_str());
     return lError;
 }
 
@@ -655,20 +706,44 @@ bool Tar::VerifyChecksum(TarHeader *Header)
 }
 
 
-//
-// class CTar, int RountTo512(int n)
-//
-// PURPOSE: Moves the number to the nearest multiple of 512.
-// RETURNS: The new number.
-//
-int Tar::RoundTo512(int n)
-{
-    if(n % 512 == 0)
-    {
-        return n;
-    }
-    else
-    {
-        return (n - (n % 512)) + 512;
+wxString Tar::GetFirstDir(wxString path) {
+    wxString firstDir = "";
+    wxString previousDir = "";
+    // tokenizer never returns empty strings + distinguish dir// and a/$
+    if (path.EndsWith("/") || path.EndsWith("\\"))
+        path += "dummy";
+    wxStringTokenizer tokenizer(path, "/\\", wxTOKEN_STRTOK);
+    while (tokenizer.HasMoreTokens()) {
+        wxString curDir = tokenizer.GetNextToken();
+        if (curDir == '.')
+	    continue;
+        if (previousDir != "") {
+	  firstDir = previousDir;
+	  break;
+	}
+        previousDir = curDir;
     }
+    return firstDir;
+}
+
+// Security: check if archive tries to jump out of destination directory
+bool Tar::IsPathInsecure(wxString path) {
+    // Avoid leading slashes (even if we preprend destdir)
+    if (path[0] == '/' || path[0] == '\\')
+        return true;
+    // Avoid ':' since wxFileName::Mkdir silently normalizes
+    // e.g. C:\test1\C:\test2 to C:\test2
+    if (path.Contains(":"))
+        return true;
+    // Ensure all files reside in the same subdirectory
+    if (GetFirstDir(path) != mInstalledDmodDirectory)
+        return true;
+    // Ensure there's no '..' path element
+    wxStringTokenizer tokenizer(path, "/\\");
+    while (tokenizer.HasMoreTokens()) {
+        wxString token = tokenizer.GetNextToken();
+        if (token == "..")
+            return true;
+    }
+    return false;
 }
